# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file defines the `HtmlUtils` classes and utility functions
"""

from .utils import system_time_millis


class HtmlUtils:
    """ Utility class for formatting code as HTML and other notebook related formatting

    """

    def __init__(self):
        pass

    @classmethod
    def formatCodeAsHtml(cls, codeText):
        """ Formats supplied code as Html suitable for use with notebook ``displayHTML``

        :param codeText: Code to be wrapped in html section
        :return: Html string

        This will wrap the code with a html section using html ``pre`` and ``code`` tags.

        It adds a copy text to clipboard button to enable users to easily copy the code to the clipboard.

        It does not reformat code so supplied code should be preformatted into lines.

        .. note::
            As the notebook environment uses IFrames in rendering html within ``displayHtml``, it cannot use
            the newer ``navigator`` based functionality as this is blocked for cross domain IFrames by default.

        """
        ts = system_time_millis()

        formattedCode = f"""
            <h3>Generated Code</h3>
            <div style="outline: 1px dashed blue;"><p ><pre><code id="generated_code_{ts}"> 
              {codeText}
            </code></pre></p></br>
            </div>
            <p><button type="button" onclick="dbldatagen_copy_code_to_clipboard()">Copy code to clipboard!</button> </p>
            <script>
            function dbldatagen_copy_code_to_clipboard() {{
               try {{
                 var r = document.createRange();
                 r.selectNode(document.getElementById("generated_code_{ts}"));
                 window.getSelection().removeAllRanges();
                 window.getSelection().addRange(r);
                 document.execCommand('copy');
                 window.getSelection().removeAllRanges();
               }}
               catch {{
                 console.error("copy to clipboard failed")
               }}
            }}
        </script>
        """

        return formattedCode

    @classmethod
    def formatTextAsHtml(cls, textContent, title="Output"):
        """ Formats supplied text as Html suitable for use with notebook ``displayHTML``

        :param textContent: Text to be wrapped in html section
        :param title: Title text to be used
        :return: Html string

        This will wrap the text content with with Html formatting

        """
        ts = system_time_millis()
        formattedContent = f"""
            <h3>{title}</h3>
            <div style="outline: 1px dashed blue;"><p ><pre id="generated_content_{ts}"> 
              {textContent}
            </pre></p></br>
            </div>
            <p><button type="button" onclick="dbldatagen_copy_to_clipboard()">Copy output to clipboard!</button></p>
            <script>
            function dbldatagen_copy_to_clipboard() {{
               try {{
                 var r = document.createRange();
                 r.selectNode(document.getElementById("generated_content_{ts}"));
                 window.getSelection().removeAllRanges();
                 window.getSelection().addRange(r);
                 document.execCommand('copy');
                 window.getSelection().removeAllRanges();
               }}
               catch {{
                 console.error("copy to clipboard failed")
               }}
            }}
        </script>
        """

        return formattedContent



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This module defines the DataRange abstract class

its not used directly, but used as base type for explicit DateRange and NRange types to ensure correct tracking of
changes to method names when refactoring

"""


class DataRange(object):
    """ Abstract class used as base class for NRange and DateRange """

    def isEmpty(self):
        """Check if object is empty (i.e all instance vars of note are `None`)"""
        raise NotImplementedError("method not implemented")

    def isFullyPopulated(self):
        """Check is all instance vars are populated"""
        raise NotImplementedError("method not implemented")

    def adjustForColumnDatatype(self, ctype):
        """ Adjust default values for column output type"""
        raise NotImplementedError("method not implemented")

    def getDiscreteRange(self):
        """Convert range to discrete range"""
        raise NotImplementedError("method not implemented")

    def getContinuousRange(self):
        """Convert range to continuous range"""
        raise NotImplementedError("method not implemented")

    def getScale(self):
        """Get scale of range"""
        raise NotImplementedError("method not implemented")

    @property
    def min(self):
        """get the `min` attribute"""
        return self.minValue
        
    @property
    def max(self):
        """get the `max` attribute"""
        return self.maxValue



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
.. title::Column Spec Options

This file defines the `ColumnSpecOptions` class
"""

import copy

from .utils import ensure


class ColumnSpecOptions(object):
    """ Column spec options object - manages options for column specs.

    This class has limited functionality - mainly used to validate and document the options,
    and the class is meant for internal use only.

    :param props: Used to pass list of properties for column generation spec property checking.

    The following options are permitted on data generator `withColumn`, `withColumnSpec` and `withColumnSpecs` methods:

    :param name: Column name

    :param type: Data type of column. Can be either instance of Spark SQL Datatype such as `IntegerType()` \
                 or string containing SQL name of type

    :param minValue: Minimum value for range of generated value. \
                     As an alternative, you may use the `dataRange` parameter

    :param maxValue: Maximum value for range of generated value. \
                     As an alternative, you may use the `dataRange` parameter

    :param step: Step to use for range of generated value. As an alternative, you may use the `dataRange` parameter

    :param numColumns: generate `n` columns numbered from 1 .. n-1 with same definition

    :param numFeatures: generate `n` columns numbered from 0 .. n-1 with same definition. Alias for `numColumns`

    :param structType: If specified as "array" and used with numColumns / numFeatures, will combine columns as array

    :param random: If True, will generate random values for column value. Defaults to `False`

    :param baseColumn: Either the string name of the base column, or a list of columns to use to
                        control data generation. The option ``baseColumns`` is an alias for ``baseColumn``.

    :param values: List of discrete values for the colummn. Discrete values for the column can be strings, numbers
                   or constants conforming to type of column

    :param weights: List of discrete weights for the colummn. Should be integer values.
                    For example, you might declare a column for status values with a weighted distribution with
                    the following statement:
                    `withColumn("status", StringType(), values=['online', 'offline', 'unknown'], weights=[3,2,1])`

    :param percentNulls: Specifies numeric percentage of generated values to be populated with SQL `null`.
                          Value is fraction representing percentage between 0.0 and 1.0
                          For example: `percentNulls=0.12` will give approximately 12% nulls for this field in the
                          output.

    :param uniqueValues: Number of unique values for column.
                          If the unique values are specified for a timestamp or date field, the values will be chosen
                          working back from the end of the previous month,
                          unless `begin`, `end` and `interval` parameters are specified

    :param begin: Beginning of range for date and timestamp fields.
                   For dates and timestamp fields, use the `begin`, `end` and `interval`
                   or `dataRange` parameters instead of `minValue`, `maxValue` and `step`

    :param end: End of range for date and timestamp fields.
                   For dates and timestamp fields, use the `begin`, `end` and `interval`
                   or `dataRange` parameters instead of `minValue`, `maxValue` and `step`

    :param interval: Interval of range for date and timestamp fields.
                   For dates and timestamp fields, use the `begin`, `end` and `interval`
                   or `dataRange` parameters instead of `minValue`, `maxValue` and `step`

    :param dataRange: An instance of an `NRange` or `DateRange` object. This can be used in place of `minValue`,
                       `maxValue`, `step` or `begin`, `end`, `interval`.

    :param template: template controlling how text should be generated

    :param textSeparator: string specifying separator to be used when constructing strings with prefix and suffix

    :param prefix: string specifying prefix text to construct field from prefix and numeric value. Both `prefix` and
                   `suffix` can be used together

    :param suffix: string specifying suffix text to construct field from suffix and numeric value. Both `prefix` and
                   `suffix` can be used together

    :param omit: if True, column is omitted from the output. Used to use column for interim effect only.

    :param expr: SQL expression to control data generation. Ignores column base value if present.

    :param implicit: Used by system to mark that column has been inferred from a schema.
                     Allows definition to be explicitly overridden.

    :param precision: Used for rounding to specific decimal layout.

    :param scale: Used for rounding to specific decimal layout.

    :param distribution: Distribution for random number. Ignored if column is not random.

    :param escapeSpecialChars: if True, require escape for all special chars in template

    .. note::
        If the `dataRange` parameter is specified as well as the `minValue`, `maxValue` or `step`,
        the results are undetermined.

        For more information, see :doc:`/reference/api/dbldatagen.daterange`
        or :doc:`/reference/api/dbldatagen.nrange`.

    """

    #: the set of attributes that must be present for any columns
    _REQUIRED_PROPERTIES = {'name', 'type'}

    _PROPERTY_ALIASES = {
        'data_range': 'dataRange',
        'base_column': 'baseColumn',
        'base_column_type': 'baseColumnType',
        'base_columns': 'baseColumn',
        'baseColumns': 'baseColumn',
        'percent_nulls': 'percentNulls',
        'unique_values': 'uniqueValues',
        'random_seed_method': 'randomSeedMethod',
        'random_seed': 'randomSeed',
        'text_separator': 'textSeparator',

    }
    #: the set of attributes that are permitted for any call to data generator `withColumn` or `withColumnSpec`
    _ALLOWED_PROPERTIES = {'name', 'type', 'minValue', 'maxValue', 'step',
                           'prefix', 'random', 'distribution',
                           'range', 'baseColumn', 'baseColumnType', 'values',
                           'numColumns', 'numFeatures', 'structType',
                           'begin', 'end', 'interval', 'expr', 'omit',
                           'weights', 'description', 'continuous',
                           'percentNulls', 'template', 'format',
                           'uniqueValues', 'dataRange', 'text',
                           'precision', 'scale',
                           'randomSeedMethod', 'randomSeed',
                           'nullable', 'implicit', 'escapeSpecialChars',
                           'suffix', 'textSeparator'
                           }

    #: the set of disallowed column attributes for any call to data generator `withColumn` or `withColumnSpec`
    _FORBIDDEN_PROPERTIES = {
        'range'
    }

    #: maxValue values for each column type, only if where value is intentionally restricted
    _MAX_TYPE_RANGE = {
        'byte': 256,
        'short': 65536,
        'int': 4294967296
    }

    def __init__(self, props, aliases=None):  # TODO: check if additional options are needed here as `**kwArgs`
        self._options = props

        self._aliases = aliases if aliases is not None else self._PROPERTY_ALIASES

        # translate aliases
        # need to copy options dictionary as you cant directly change a
        # dictionary that you are iterating over
        updated_options = copy.copy(props)
        for k in props.keys():
            if k in self._aliases:
                v = props[k]
                alias_name = self._aliases[k]
                updated_options[alias_name] = v
                del updated_options[k]

        self._options = updated_options

    @property
    def options(self):
        """ Get options dictionary for object

            :return: options dictionary for object

        """
        return self._options

    def getOrElse(self, key, default=None):
        """ Get val for key if it exists or else return default"""
        assert key is not None, "key must be valid key string"

        if key in self._options:
            return self._options.get(key, default)
        if key in self._aliases:
            return self._options.get(self._aliases[key], default)
        return default

    def __getitem__(self, key):
        """ implement the built in dereference by key behavior """
        ensure(key is not None, "key should be non-empty")
        return self._options.get(key, None)

    def checkBoolOption(self, v, name=None, optional=True):
        """ Check that option is either not specified or of type boolean

        :param v: value to test
        :param name: name of value to use in any reported errors or exceptions
        :param optional: If True (default), indicates that value is optional and
                         that `None` is a valid value for the option
        """
        assert name is not None, "`name` must be specified"
        if optional:
            ensure(v is None or type(v) is bool,
                   f"Option `{name}` must be boolean if specified - value: {v}, type: {type(v)}")
        else:
            ensure(type(v) is bool,
                   f"Option `{name}` must be boolean  - value: {v}, type: {type(v)}")

    def checkExclusiveOptions(self, options):
        """check if the options are exclusive - i.e only one is not None

        :param options: list of options that will be mutually exclusive
        """
        assert options is not None, "options must be non empty"
        assert type(options) is list, "`options` must be list"
        assert len([self[x] for x in options if self[x] is not None]) <= 1, \
            f" only one of of the options: {options} may be specified "

    def checkOptionValues(self, option, option_values):
        """check if option value is in list of values

        :param option: list of options that will be mutually exclusive
        :param option_values: list of possible option values that will be mutually exclusive
        """
        assert option is not None and len(option.strip()) > 0, "option must be non empty"
        assert type(option_values) is list, "`option_values` must be list"
        assert self[option] in option_values, f"option: `{option}` must have one of the values {option_values}"

    def checkValidColumnProperties(self, columnProps):
        """
            check that column definition properties are recognized
            and that the column definition has required properties

            :param columnProps:
        """
        ensure(columnProps is not None, "columnProps should be non-empty")

        col_type = self['type']
        if col_type.typeName() in self._MAX_TYPE_RANGE:
            minValue = self['minValue']
            maxValue = self['maxValue']

            if minValue is not None and maxValue is not None:
                effective_range = maxValue - minValue
                if effective_range > self._MAX_TYPE_RANGE[col_type.typeName()]:
                    raise ValueError("Effective range greater than range of type")

        for k in columnProps.keys():
            ensure(k in ColumnSpecOptions._ALLOWED_PROPERTIES or k in ColumnSpecOptions._PROPERTY_ALIASES,
                   f"invalid column option {k}")

        for arg in self._REQUIRED_PROPERTIES:
            ensure(columnProps.get(arg) is not None, f"missing column option {arg}")

        for arg in self._FORBIDDEN_PROPERTIES:
            ensure(arg not in columnProps, f"forbidden column option {arg}")

        # check weights and values
        if 'weights' in columnProps:
            ensure('values' in columnProps,
                   f"weights are only allowed for columns with values - column '{columnProps['name']}' ")
            ensure(columnProps['values'] is not None and len(columnProps['values']) > 0,
                   f"weights must be associated with non-empty list of values - column '{columnProps['name']}' ")
            ensure(len(columnProps['values']) == len(columnProps['weights']),
                   f"length(list of weights) must equal length(list of values) - column '{columnProps['name']}' ")



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file defines the Gamma statistical distributions related classes

"""

import numpy as np
import pandas as pd
import pyspark.sql.functions as F
from pyspark.sql.types import FloatType

from .data_distribution import DataDistribution


class Gamma(DataDistribution):
    """ Specify Gamma distribution with specific shape and scale

    :param shape: shape parameter (k)
    :param scale: scale parameter (theta)

    See https://en.wikipedia.org/wiki/Gamma_distribution

    Scaling is performed to normalize values between 0 and 1

    """

    def __init__(self, shape, scale):
        DataDistribution.__init__(self)
        assert type(shape) in [float, int, np.float64, np.int32, np.int64], "alpha must be int-like or float-like"
        assert type(scale) in [float, int, np.float64, np.int32, np.int64], "beta must be int-like or float-like"
        self._shape = shape
        self._scale = scale

    @property
    def shape(self):
        """ Return shape parameter."""
        return self._shape

    @property
    def scale(self):
        """ Return scale parameter."""
        return self._scale

    def __str__(self):
        """ Return string representation of object """
        return f"GammaDistribution(shape(`k`)={self._shape}, scale(`theta`)={self._scale}, seed={self.randomSeed})"

    @staticmethod
    def gamma_func(shape_series: pd.Series, scale_series: pd.Series, random_seed: pd.Series) -> pd.Series:
        """ Pandas / Numpy based function to generate gamma samples

        :param shape_series: pandas series of shape (k) values
        :param scale_series: pandas series of scale (theta) values
        :param random_seed:  pandas series of random seed values

        :return: Samples scaled from 0 .. 1
        """
        shape = shape_series.to_numpy()
        scale = scale_series.to_numpy()
        random_seed = random_seed.to_numpy()[0]

        rng = DataDistribution.get_np_random_generator(random_seed)

        results = rng.gamma(shape, scale)

        # scale results to range [0, 1]
        amin = np.amin(results) * 1.0
        amax = np.amax(results) * 1.0

        adjusted_results = results - amin

        scaling_factor = amax - amin

        results2 = adjusted_results / scaling_factor
        return pd.Series(results2)

    def generateNormalizedDistributionSample(self):
        """ Generate sample of data for distribution

        :return: random samples from distribution scaled to values between 0 and 1
        """
        gamma_sample = F.pandas_udf(self.gamma_func, returnType=FloatType()).asNondeterministic()

        newDef = gamma_sample(F.lit(self._shape),
                             F.lit(self._scale),
                             F.lit(self.randomSeed) if self.randomSeed is not None else F.lit(-1.0))
        return newDef



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file defines the text generator plugin class `PyfuncText`
"""

import importlib
import logging

from .text_generators import TextGenerator
from .utils import DataGenError


class PyfuncText(TextGenerator):  # lgtm [py/missing-equals]
    """ Text generator that supports generating text from arbitrary Python function

    :param fn: function to call to generate text.
    :param init: function to call to initialize context
    :param initPerBatch: if init per batch is set to True, initialization of context is performed on every Pandas udf
                         call. Default is False.
    :param name: String representing name of text generator when converted to string via ``repr`` or ``str``

    The two functions define the plugin model

    The first function, ``fn`` is called whenever text should be generated for a single column of a single row

    It is called with the signature ``fn(context, value)`` unless a root property is set, in which the signature is
    ``fn(rootProperty)`` with rootProperty having the value of the root property of the context.

    Context is the stored context containing instances of random number generators, 3rd party
    client library objects etc.

    The ``initFn`` is called to initialize the function call context. The plugin code can store arbitrary properties
    in the context following normal Python object rules.

    The context is initialized with the property `textGenerator` prior to being initialized which is a reference to the
    enclosing text generator.

    .. note::
              There are no expectations of repeatability of data generation when using external code
              or external libraries to generate text.

              However, custom code can call the base class method to get a Numpy random
              number generator instance. This will have been seeded using the ``dbldatagen``
              random number seed if one was specified, so random numbers generated from this will be repeatable.

              The custom code may call the property ``randomSeed`` on the text generator object to get the random seed
              which may be used to seed library specific initialization.

              This random seed property may have the values ``None`` or ``-1`` which should be treated as meaning dont
              use a random seed.

              The code does not guarantee thread or cross process safety. If a new instance of the random number
              generator is needed, you may call the base class method with the argument `forceNewInstance` set to True.
    """

    class _FnCallContext:
        """ inner class to support storage of context between calls

            initial instances of random number generators, clients for services etc here during execution
            of the `initFn` calls

            :param txtGen: - reference to outer PyfnText object

        """

        def __init__(self, txtGen):
            self.textGenerator = txtGen

    def __init__(self, fn, init=None, initPerBatch=False, name=None, rootProperty=None):
        super().__init__()
        assert fn is not None or callable(fn), "Function must be provided wiith signature fn(context, oldValue)"
        assert init is None or callable(init), "Init function must be a callable function or lambda if passed"

        # if root property is provided, root property will be passed to generate text function
        self._rootProperty = rootProperty

        self._pyFn = fn  # generate text function
        self._initFn = init  # context initialization function
        self._context = None  # context used to hold library root object and other properties

        # if init per batch is True, initialization of context will be per UDF call
        assert initPerBatch in [True, False], "initPerBatch must evaluate to boolean True or False"
        self._initPerBatch = initPerBatch

        self._name = name if name is not None else "PyfuncText"

    def __str__(self):
        """ Get string representation of object
            ``name`` property is used to provide user friendly name for text generator
        """
        return f"{self._name}({repr(self._pyFn)}, init={self._initFn})"

    def _getContext(self, forceNewInstance=False):
        """ Get the context for plugin function calls

        :param forceNewInstance: if True, forces each call to create a new context
        :return: existing or newly created context.

        """
        context = self._context
        if context is None or forceNewInstance:
            context = PyfuncText._FnCallContext(self)

            # init context using context creator if any provided
            if self._initFn is not None:
                self._initFn(context)

            # save context for later use unless forceNewInstance is set
            if not forceNewInstance:
                self._context = context
            else:
                return context
        return self._context

    def pandasGenerateText(self, v):
        """ Called to generate text via Pandas UDF mechanism

        :param v: base value of column as Pandas Series

        """
        # save object properties in local vars to avoid overhead of object dereferences
        # on every call
        context = self._getContext(self._initPerBatch)
        evalFn = self._pyFn
        rootProperty = getattr(context, self._rootProperty) if self._rootProperty is not None else None

        # define functions to call with context and with root property
        def _valueFromFn(originalValue):
            return evalFn(context, originalValue)

        def _valueFromFnWithRoot(originalValue):
            return evalFn(rootProperty)

        if rootProperty is not None:
            results = v.apply(_valueFromFnWithRoot, args=None)
        else:
            results = v.apply(_valueFromFn, args=None)

        return results


class PyfuncTextFactory:
    """PyfuncTextFactory applies syntactic wrapping around creation of PyfuncText objects

    :param name: name of generated object (when converted to string via ``str``)

    It allows the use of the following constructs:

    .. code-block:: python

       # initialization (for Faker for example)

       # setup use of Faker
       def initFaker(ctx):
         ctx.faker = Faker(locale="en_US")
         ctx.faker.add_provider(internet)

       FakerText = (PyfuncTextFactory(name="FakerText")
                   .withInit(initFaker)        # determines how context should be initialized
                   .withRootProperty("faker")  # determines what context property is passed to fn
                   )

       # later use ...
       .withColumn("fake_name", text=FakerText("sentence", ext_word_list=my_word_list) )
       .withColumn("fake_sentence", text=FakerText("sentence", ext_word_list=my_word_list) )

       # translates to generation of lambda function with keyword arguments
       # or without as needed
       .withColumn("fake_name",
                   text=FakerText( (lambda faker: faker.name( )),
                                   init=initFaker,
                                   rootProperty="faker",
                                   name="FakerText"))
       .withColumn("fake_sentence",
                   text=FakerText( (lambda faker:
                                       faker.sentence( **{ "ext_word_list" : my_word_list} )),
                                   init=initFaker,
                                   rootProperty="faker",
                                   name="FakerText"))

    """

    def __init__(self, name=None):
        """

        :param name: name of generated object (when converted to string via ``str``)

        """
        self._initFn = None
        self._rootProperty = None
        self._name = "PyfuncText" if name is None else name
        self._initPerBatch = False

    def withInit(self, fn):
        """ Specifies context initialization function

            :param fn: function pointer or lambda function for initialization
                       signature should ``initFunction(context)``

            .. note::
               This variation initializes the context once per worker process per text generator
               instance.
        """
        self._initFn = fn
        return self

    def withInitPerBatch(self, fn):
        """ Specifies context initialization function

            :param fn: function pointer or lambda function for initialization
                       signature should ``initFunction(context)``

            .. note::
               This variation initializes the context once per internal pandas UDF call.
               The UDF call will be called once per 10,000 rows if system is configured using defaults.
               Setting the pandas batch size as an argument to the DataSpec creation will change the default
               batch size.
        """
        self._initPerBatch = True
        return self.withInit(fn)

    def withRootProperty(self, prop):
        """ If called, specifies the property of the context to be passed to the text generation function.
            If not called, the context object itself will be passed to the text generation function.
        """
        self._rootProperty = prop
        return self

    def __call__(self, evalFn, *args, isProperty=False, **kwargs):
        """ Internal function call mechanism that implements the syntax expansion

        :param evalFn: text generation function or lambda
        :param args: optional args to be passed by position
        :param kwargs: optional keyword args following Python keyword passing mechanism
        :param isProperty: if true, interpret evalFn as string name of property, not a function or method
        """
        assert evalFn is not None and (type(evalFn) is str or callable(evalFn)), "Function must be provided"

        if type(evalFn) is str:
            assert self._rootProperty is not None and len(self._rootProperty.strip()) > 0, \
                "string named functions can only be used on text generators with root property"
            fnName = evalFn
            if len(args) > 0 and len(kwargs) > 0:
                # generate lambda with both kwargs and args
                assert not isProperty, "isProperty cannot be true if using arguments"
                evalFn = lambda root: getattr(root, fnName)(*args, **kwargs)
            elif len(args) > 0:
                # generate lambda with positional args
                assert not isProperty, "isProperty cannot be true if using arguments"
                evalFn = lambda root: getattr(root, fnName)(*args)
            elif len(kwargs) > 0:
                # generate lambda with keyword args
                assert not isProperty, "isProperty cannot be true if using arguments"
                evalFn = lambda root: getattr(root, fnName)(**kwargs)
            elif isProperty:
                # generate lambda with property access, not method call
                evalFn = lambda root: getattr(root, fnName)
            else:
                # generate lambda with no args
                evalFn = (lambda root: getattr(root, fnName)())

        # returns the actual PyfuncText text generator object.
        # Note all syntax expansion is performed once only
        return PyfuncText(evalFn, init=self._initFn, name=self._name, rootProperty=self._rootProperty)


class FakerTextFactory(PyfuncTextFactory):
    """ Factory object for Faker text generator flavored ``PyfuncText`` objects

    :param locale: list of locales. If empty, defaults to ``en-US``
    :param providers: list of providers
    :param name: name of generated objects. Defaults to ``FakerText``
    :param lib: library import name of Faker library. If none passed, uses ``faker``
    :param rootClass: name of root object class If none passed, uses ``Faker``

    ..note ::
       Both the library name and root object class can be overridden - this is primarily for internal testing purposes.
    """

    _FAKER_LIB = "faker"

    _defaultFakerTextFactory = None

    def __init__(self, locale=None, providers=None, name="FakerText", lib=None,
                 rootClass=None):

        super().__init__(name)

        # set up the logger
        self._logger = logging.getLogger("FakerTextFactory")
        self._logger.setLevel(logging.WARNING)

        # setup Faker library to use
        if lib is None:
            lib = self._FAKER_LIB

        # allow overriding the root object class for test purposes
        if rootClass is None:
            self._rootObjectClass = "Faker"
        else:
            self._rootObjectClass = rootClass

        # load the library
        fakerModule = self._loadLibrary(lib)

        # make the initialization function
        initFn = self._mkInitFn(fakerModule, locale, providers)

        self.withInit(initFn)
        self.withRootProperty("faker")

    @classmethod
    def _getDefaultFactory(cls, lib=None, rootClass=None):
        """Class method to get default faker text factory

           Not intended for general use
        """
        if cls._defaultFakerTextFactory is None:
            cls._defaultFakerTextFactory = FakerTextFactory(lib=lib, rootClass=rootClass)
        return cls._defaultFakerTextFactory

    def _mkInitFn(self, libModule, locale, providers):
        """ Make Faker initialization function

        :param locale: locale string or list of locale strings
        :param providers: providers to load
        :return:
        """
        assert libModule is not None, "must have a valid loaded Faker library module"

        fakerClass = getattr(libModule, self._rootObjectClass)

        # define the initialization function for Faker
        def fakerInitFn(ctx):
            if locale is not None:
                ctx.faker = fakerClass(locale=locale)
            else:
                ctx.faker = fakerClass()

            if providers is not None:
                for provider in providers:
                    ctx.faker.add_provider(provider)

        return fakerInitFn

    def _loadLibrary(self, lib):
        """ Load faker library if not already loaded

        :param lib: library name of Faker library. If none passed, uses ``faker``
        """
        # load library
        try:
            if lib is not None:
                assert type(lib) is str and len(lib.strip()), f"Library ``{lib}`` must be a valid library name"

                if lib in globals():
                    return globals()[lib]
                else:
                    fakerModule = importlib.import_module(lib)
                    globals()[lib] = fakerModule
                    return fakerModule
        except RuntimeError as err:
            # pylint: disable=raise-missing-from
            raise DataGenError("Could not load or initialize Faker library", err)


def fakerText(mname, *args, _lib=None, _rootClass=None, **kwargs):
    """Generate faker text generator object using default FakerTextFactory
       instance

       :param mname: method name to invoke
       :param args: positional args to be passed to underlying Faker instance
       :param _lib: internal only param - library to load
       :param _rootClass: internal only param - root class to create
       
       :returns : instance of PyfuncText for use with Faker

       ``fakerText("sentence")`` is same as ``FakerTextFactory()("sentence")``
    """
    defaultFactory = FakerTextFactory._getDefaultFactory(lib=_lib,
                                                         rootClass=_rootClass)
    return defaultFactory(mname, *args, **kwargs)  # pylint: disable=not-callable



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file defines the `ColumnGenerationSpec` class
"""

import copy
import logging

from pyspark.sql.functions import col, pandas_udf
from pyspark.sql.functions import lit, concat, rand, round as sql_round, array, expr, when, udf, \
    format_string

import pyspark.sql.functions as F

from pyspark.sql.types import FloatType, IntegerType, StringType, DoubleType, BooleanType, \
    TimestampType, DataType, DateType, ArrayType, MapType, StructType

from .column_spec_options import ColumnSpecOptions
from .datagen_constants import RANDOM_SEED_FIXED, RANDOM_SEED_HASH_FIELD_NAME, RANDOM_SEED_RANDOM, \
    DEFAULT_SEED_COLUMN, OPTION_RANDOM, OPTION_RANDOM_SEED, OPTION_RANDOM_SEED_METHOD, INFER_DATATYPE

from .daterange import DateRange
from .distributions import Normal, DataDistribution
from .nrange import NRange
from .text_generators import TemplateGenerator
from .utils import ensure, coalesce_values
from .schema_parser import SchemaParser

HASH_COMPUTE_METHOD = "hash"
VALUES_COMPUTE_METHOD = "values"
RAW_VALUES_COMPUTE_METHOD = "raw_values"
AUTO_COMPUTE_METHOD = "auto"
EXPR_OPTION = "expr"
COMPUTE_METHOD_VALID_VALUES = [HASH_COMPUTE_METHOD,
                               AUTO_COMPUTE_METHOD,
                               VALUES_COMPUTE_METHOD,
                               RAW_VALUES_COMPUTE_METHOD]


class ColumnGenerationSpec(object):
    """ Column generation spec object - specifies how column is to be generated

    Each column to be output will have a corresponding ColumnGenerationSpec object.
    This is added explicitly using the DataGenerators `withColumnSpec` or `withColumn` methods

    If none is explicitly added, a default one will be generated.

    The full set of arguments to the class is more than the explicitly called out parameters as any
    arguments that are not explicitly called out can still be passed due to the `**kwargs` expression.

    This class is meant for internal use only.

    :param name: Name of column (string).
    :param colType: Spark SQL datatype instance, representing the type of the column.
    :param min: minimum value of column
    :param max: maximum value of the column
    :param step: numeric step used in column data generation
    :param prefix: string used as prefix to the column underlying value to produce a string value
    :param random: Boolean, if True, will generate random values
    :param distribution: Instance of distribution, that will control the distribution of the generated values
    :param baseColumn: String or list of strings representing columns used as basis for generating the column data
    :param randomSeed: random seed value used to generate the random value, if column data is random
    :param randomSeedMethod: method for computing random values from the random seed. It may take on the
           values `fixed`, `hash_fieldname` or None

    :param implicit: If True, the specification for the column can be replaced by a later definition.
           If not, a later attempt to replace the definition will flag an error.
           Typically used when generating definitions automatically from a schema, or when using wildcards
           in the specification

    :param omit: if True, omit from the final output.
    :param nullable: If True, column may be null - defaults to True.
    :param debug: If True, output debugging log statements. Defaults to False.
    :param verbose: If True, output logging statements at the info level. If False (the default),
                    only output warning and error logging statements.
    :param seedColumnName: if supplied, specifies seed column name

    For full list of options, see :doc:`/reference/api/dbldatagen.column_spec_options`.
    """

    #: maxValue values for each column type, only if where value is intentionally restricted
    _max_type_range = {
        'byte': 256,
        'short': 65536
    }

    _ARRAY_STRUCT_TYPE = "array"

    # set up logging

    # restrict spurious messages from java gateway
    logging.getLogger("py4j").setLevel(logging.WARNING)

    def __init__(self, name, colType=None, minValue=0, maxValue=None, step=1, prefix='', random=False,
                 distribution=None, baseColumn=None, randomSeed=None, randomSeedMethod=None,
                 implicit=False, omit=False, nullable=True, debug=False, verbose=False,
                 seedColumnName=DEFAULT_SEED_COLUMN,
                 **kwargs):

        # set up logging
        self.verbose = verbose
        self.debug = debug

        self._setup_logger()

        # set up default range and type for column
        self._dataRange = NRange(None, None, None)  # by default range of values for  column is unconstrained

        self._inferDataType = False
        if colType is None:  # default to integer field if none specified
            colType = IntegerType()
        elif colType == INFER_DATATYPE:
            colType = StringType()  # default inferred data type to string until exact type is known
            self._inferDataType = True

            if EXPR_OPTION not in kwargs:
                raise ValueError("Column generation spec must have `expr` attribute specified if datatype is inferred")

        elif type(colType) == str:
            colType = SchemaParser.columnTypeFromString(colType)

        assert isinstance(colType, DataType), f"colType `{colType}` is not instance of DataType"

        self._initialBuildPlan = []  # the build plan for the column - descriptive only
        self.executionHistory = []  # the execution history for the column

        self._seedColumnName = seedColumnName

        # If no base column is specified, assume its dependent on the seed column
        if baseColumn is None:
            baseColumn = self._seedColumnName

        # to allow for open ended extension of many column attributes, we use a few specific
        # parameters and pass the rest as keyword arguments
        supplied_options = {'name': name, 'minValue': minValue, 'type': colType,
                            'maxValue': maxValue, 'step': step,
                            'prefix': prefix, 'baseColumn': baseColumn,
                            OPTION_RANDOM: random, 'distribution': distribution,
                            OPTION_RANDOM_SEED_METHOD: randomSeedMethod, OPTION_RANDOM_SEED: randomSeed,
                            'omit': omit, 'nullable': nullable, 'implicit': implicit
                            }

        supplied_options.update(kwargs)

        self._csOptions = ColumnSpecOptions(supplied_options)

        self._csOptions.checkValidColumnProperties(supplied_options)

        # only allow `template` or `text`
        self._csOptions.checkExclusiveOptions(["template", "text"])

        # only allow `weights` or `distribution`
        self._csOptions.checkExclusiveOptions(["distribution", "weights"])

        # check for alternative forms of specifying range
        # column_spec_options._checkExclusiveOptions(["minValue", "minValue", "begin", "dataRange"])
        # column_spec_options._checkExclusiveOptions(["maxValue", "maxValue", "end", "dataRange"])
        # column_spec_options._checkExclusiveOptions(["step", "interval", "dataRange"])

        # we want to assign each of the properties to the appropriate instance variables
        # but compute sensible defaults in the process as needed
        # in particular, we want to ensure that things like values and weights match
        # and that minValue and maxValue are not inconsistent with distributions, ranges etc

        # if a column spec is implicit, it can be overwritten
        # by default column specs added by wild cards or inferred from schemas are implicit
        self._csOptions.checkBoolOption(implicit, name="implicit")
        self.implicit = implicit

        # if true, omit the column from the final output

        self._csOptions.checkBoolOption(omit, name="omit")
        self.omit = omit

        # the column name
        self.name = name

        # the base column data types
        self._baseColumnDatatypes = []

        # not used for much other than to validate against option to generate nulls
        self._csOptions.checkBoolOption(nullable, name="nullable")
        self.nullable = nullable

        # should be either a literal or None
        # use of a random seed method will ensure that we have repeatability of data generation
        assert randomSeed is None or type(randomSeed) in [int, float], "seed should be None or numeric"

        assert randomSeedMethod is None or randomSeedMethod in [RANDOM_SEED_FIXED, RANDOM_SEED_HASH_FIELD_NAME], \
            f"`randomSeedMethod` should be none or `{RANDOM_SEED_FIXED}` or `{RANDOM_SEED_HASH_FIELD_NAME}`"

        self._randomSeedMethod = self[OPTION_RANDOM_SEED_METHOD]
        self.random = self[OPTION_RANDOM]

        if self._randomSeedMethod == RANDOM_SEED_HASH_FIELD_NAME:
            assert self.name is not None, "field name cannot be None"
            self._randomSeed = abs(hash(self.name))
        else:
            self._randomSeed = self[OPTION_RANDOM_SEED]

        # random seed method should be "fixed" or "hash_fieldname"
        if self._randomSeed is not None and self._randomSeedMethod is None:
            self._randomSeedMethod = RANDOM_SEED_FIXED

        # compute dependencies
        self.dependencies = self._computeBasicDependencies()

        # value of `base_column_type` must be `None`,"values", "raw_values", "auto",  or "hash"
        # this is the method of computing current column value from base column, not the data type of the base column
        allowed_compute_methods = [AUTO_COMPUTE_METHOD, VALUES_COMPUTE_METHOD, HASH_COMPUTE_METHOD,
                                   RAW_VALUES_COMPUTE_METHOD, None]
        self._csOptions.checkOptionValues("baseColumnType", allowed_compute_methods)
        self._baseColumnComputeMethod = self['baseColumnType']

        # handle text generation templates
        if self['template'] is not None:
            assert isinstance(self['template'], str), "template must be a string "
            escapeSpecialChars = self['escapeSpecialChars'] if self['escapeSpecialChars'] is not None else False
            self._textGenerator = TemplateGenerator(self['template'], escapeSpecialChars)
        elif self['text'] is not None:
            self._textGenerator = copy.deepcopy(self['text'])
        else:
            self._textGenerator = None

        # specify random seed for text generator if one is in effect
        if self._textGenerator is not None and self._randomSeed is not None:
            self._textGenerator = self._textGenerator.withRandomSeed(self._randomSeed)

        # compute required temporary values
        self.temporaryColumns = []

        data_range = self["dataRange"]

        unique_values = self["uniqueValues"]

        c_min, c_max, c_step = (self["minValue"], self["maxValue"], self["step"])
        c_begin, c_end, c_interval = self['begin'], self['end'], self['interval']

        # handle weights / values and distributions
        self.weights, self.values = (self["weights"], self["values"])

        self.distribution = self["distribution"]

        # if distribution is just specified as `normal` use standard normal distribution
        if self.distribution == "normal":
            self.distribution = Normal.standardNormal()

        # specify random seed for distribution if one is in effect
        if self.distribution is not None and self._randomSeed is not None:
            self.distribution = self.distribution.withRandomSeed(self._randomSeed)

        # force weights and values to list
        if self.weights is not None:
            # coerce to list - this will allow for pandas series, numpy arrays and tuples to be used
            self.weights = list(self.weights)

        if self.values is not None:
            # coerce to list - this will allow for pandas series, numpy arrays and tuples to be used
            self.values = list(self.values)

        # handle default method of computing the base column value
        # if we have text manipulation, use 'values' as default for format but 'hash' as default if
        # its a column with multiple values
        if self._baseColumnComputeMethod in [None, AUTO_COMPUTE_METHOD] \
                and (self.textGenerator is not None or self['format'] is not None
                     or self['prefix'] is not None or self['suffix'] is not None):
            if self.values is not None:
                self.logger.info("""Column [%s] has no `base_column_type` attribute and uses discrete values
                                       => Assuming `hash` for attribute `base_column_type`. 
                                       => Use explicit value for `base_column_type` if alternate interpretation needed

                                    """, self.name)
                self._baseColumnComputeMethod = HASH_COMPUTE_METHOD
            else:
                self.logger.info("""Column [%s] has no `base_column_type` attribute specified for formatted text
                                       => Assuming `values` for attribute `base_column_type`. 
                                       => Use explicit value for `base_column_type` if alternate interpretation  needed

                                    """, self.name)
                self._baseColumnComputeMethod = VALUES_COMPUTE_METHOD

        # adjust the range by merging type and range information

        self._dataRange = self._computeAdjustedRangeForColumn(colType=colType,
                                                              c_min=c_min, c_max=c_max, c_step=c_step,
                                                              c_begin=c_begin, c_end=c_end,
                                                              c_interval=c_interval,
                                                              c_unique=unique_values, c_range=data_range)

        if self.distribution is not None:
            ensure((self._dataRange is not None and self._dataRange.isFullyPopulated())
                   or
                   self.values is not None,
                   """When using an explicit distribution, provide a fully populated range or a set of values""")

        # set up the temporary columns needed for data generation
        self._setupTemporaryColumns()

    def _temporaryRename(self, tmpName):
        """ Create enter / exit object to support temporary renaming of column spec

        This is to support the functionality:

        ```
           with columSpec._temporaryRename("test") as modifiedColumn:
             modifiedColumn.doSomthing()
        ```

        When building array or multi-column valued columns, we rename the column definition temporarily
        to ensure that logging, saving messages to the execution plan and random number generation based on
        the field name work correctly

        :param tmpName: temporary name for the column.
        :return: object supporting `with columSpec._temporaryRename("test") as modifiedColumn:` semantics

        .. note::
           This does not create a copy of the column spec object

        """
        # create class for temporary rename enter / exit
        assert tmpName is not None and len(tmpName) > 0, "method expects valid temporary name"

        class RenameEnterExit:
            def __init__(self, columnSpec, newName):
                """ Save column spec and old name to support enter / exit semantics """
                self._cs = columnSpec
                self._oldName = columnSpec.name
                self._newName = newName
                self._randomSeed = columnSpec._randomSeed

            def __enter__(self):
                """ Return the inner column spec object """
                self._cs.name = self._newName

                if self._cs._randomSeedMethod == RANDOM_SEED_HASH_FIELD_NAME:
                    self._cs._randomSeed = abs(hash(self._cs.name))

                    if self._cs._textGenerator is not None and self._cs._randomSeed is not None:
                        self._cs._textGenerator = self._cs._textGenerator.withRandomSeed(self._cs._randomSeed)

                return self._cs

            def __exit__(self, exc_type, exc_value, tb):
                # restore old name
                self._cs.name = self._oldName

                # restore old random seed
                self._cs._randomSeed = self._randomSeed

                if self._cs._randomSeedMethod == RANDOM_SEED_HASH_FIELD_NAME:
                    if self._cs._textGenerator is not None and self._cs._randomSeed is not None:
                        self._cs._textGenerator = self._cs._textGenerator.withRandomSeed(self._cs._randomSeed)

                if exc_type is not None:
                    # uncomment for traceback
                    # import traceback
                    # traceback.print_exception(exc_type, exc_value, tb)
                    return False

                return True

        return RenameEnterExit(self, tmpName)

    @property
    def specOptions(self):
        """ get column spec options for spec

        .. note::
            This is intended for testing use only.
            Option values set directly through the options dict are not supported.

        :return: underlying options object
        """
        return self._csOptions.options

    def __deepcopy__(self, memo):
        """Custom deep copy method that resets the logger to avoid trying to copy the logger

        :see https://docs.python.org/3/library/copy.html
        """
        self.logger = None  # pylint: disable=attribute-defined-outside-init
        result = None

        try:
            cls = self.__class__
            result = cls.__new__(cls)
            memo[id(self)] = result
            for k, v in self.__dict__.items():
                setattr(result, k, copy.deepcopy(v, memo))
        finally:
            self._setup_logger()
            if result is not None:
                result._setup_logger()
        return result

    @property
    def randomSeed(self):
        """ get random seed for column spec"""
        return self._randomSeed

    @property
    def isRandom(self):
        """ returns True if column will be randomly generated"""
        return self[OPTION_RANDOM]

    @property
    def textGenerator(self):
        """ Get the text generator for the column spec"""
        return self._textGenerator

    @property
    def inferDatatype(self):
        """ If True indicates that datatype should be inferred to be result of computing SQL expression
        """
        return self._inferDataType

    @property
    def baseColumns(self):
        """ Return base columns as list of strings"""

        # if base column  is string and contains multiple columns, split them
        # other build list of columns if needed
        if type(self.baseColumn) is str and "," in self.baseColumn:
            return [x.strip() for x in self.baseColumn.split(",")]
        elif type(self.baseColumn) is list:
            return self.baseColumn
        else:
            return [self.baseColumn]

    def _computeBasicDependencies(self):
        """ get set of basic column dependencies.

        These are used to compute the order of field evaluation

        :return: base columns as list with dependency on seed column added
        """
        if self.baseColumn != self._seedColumnName:
            return list(set(self.baseColumns + [self._seedColumnName]))
        else:
            return [self._seedColumnName]

    def setBaseColumnDatatypes(self, columnDatatypes):
        """ Set the data types for the base columns

        :param column_datatypes: = list of data types for the base columns

        """
        assert type(columnDatatypes) is list, " `column_datatypes` parameter must be list"
        ensure(len(columnDatatypes) == len(self.baseColumns),
               "number of base column datatypes must match number of  base columns")
        self._baseColumnDatatypes = [].append(columnDatatypes)

    def _setupTemporaryColumns(self):
        """ Set up any temporary columns needed for test data generation.

        For some types of test data, intermediate columns are used in the data generation process
        but dropped from the final output
        """
        if self.isWeightedValuesColumn:
            # if its a weighted values column, then create temporary for it
            # not supported for feature / array columns for now
            min_num_columns, max_num_columns, struct_type = self._getMultiColumnDetails(validate=False)
            ensure(max_num_columns is None or max_num_columns <= 1,
                   "weighted columns not supported for multi-column or multi-feature values")
            if self.random:
                temp_name = f"_rnd_{self.name}"
                self.dependencies.append(temp_name)
                desc = f"adding temporary column {temp_name} required by {self.name}"
                self._initialBuildPlan.append(desc)
                sql_random_generator = self._getUniformRandomSQLExpression(self.name)
                self.temporaryColumns.append((temp_name, DoubleType(), {'expr': sql_random_generator, 'omit': True,
                                                                        'description': desc}))
                self._weightedBaseColumn = temp_name
            else:
                # create temporary expression mapping values to range of weights
                temp_name = f"_scaled_{self.name}"
                self.dependencies.append(temp_name)
                desc = f"adding temporary column {temp_name} required by {self.name}"
                self._initialBuildPlan.append(desc)

                # use a base expression based on mapping base column to size of data
                sql_scaled_generator = self._getScaledIntSQLExpression(self.name,
                                                                       scale=sum(self.weights),
                                                                       base_columns=self.baseColumns,
                                                                       base_datatypes=self._baseColumnDatatypes,
                                                                       compute_method=self._baseColumnComputeMethod,
                                                                       normalize=True)

                self.logger.debug("""building scaled sql expression : '%s' 
                                      with base column: %s, dependencies: %s""",
                                  sql_scaled_generator,
                                  self.baseColumn,
                                  self.dependencies)

                self.temporaryColumns.append((temp_name, DoubleType(), {'expr': sql_scaled_generator, 'omit': True,
                                                                        'baseColumn': self.baseColumn,
                                                                        'description': desc}))
                self._weightedBaseColumn = temp_name

    def _setup_logger(self):
        """Set up logging

        This will set the logger at warning, info or debug levels depending on the instance construction parameters
        """
        self.logger = logging.getLogger("DataGenerator")
        if self.debug:
            self.logger.setLevel(logging.DEBUG)
        elif self.verbose:
            self.logger.setLevel(logging.INFO)
        else:
            self.logger.setLevel(logging.WARNING)

    def _computeAdjustedRangeForColumn(self, colType, c_min, c_max, c_step, c_begin, c_end, c_interval, c_range,
                                       c_unique):
        """Determine adjusted range for data column
        """
        assert colType is not None, "`colType` must be non-None instance"

        if type(colType) is DateType or type(colType) is TimestampType:
            return self._computeAdjustedDateTimeRangeForColumn(colType, c_begin, c_end, c_interval, c_range, c_unique)
        else:
            return self._computeAdjustedNumericRangeForColumn(colType, c_min, c_max, c_step, c_range, c_unique)

    def _computeAdjustedNumericRangeForColumn(self, colType, c_min, c_max, c_step, c_range, c_unique):
        """Determine adjusted range for data column

        Rules:
        - if a datarange is specified , use that range
        - if begin and end are specified or minValue and maxValue are specified, use that
        - if unique values is specified, compute minValue and maxValue depending on type

        """
        if c_unique is not None:
            assert type(c_unique) is int, "unique_values must be integer"
            assert c_unique >= 1, "if supplied, unique values must be > 0"
            # TODO: set maxValue to unique_values + minValue & add unit test
            effective_min, effective_max, effective_step = None, None, None
            if c_range is not None and type(c_range) is NRange:
                effective_min = c_range.minValue
                effective_step = c_range.step
                effective_max = c_range.maxValue
            effective_min = coalesce_values(effective_min, c_min, 1)
            effective_step = coalesce_values(effective_step, c_step, 1)
            effective_max = coalesce_values(effective_max, c_max)

            # due to floating point errors in some Python floating point calculations, we need to apply rounding
            # if any of the components are float
            if type(effective_min) is float or type(effective_step) is float:
                unique_max = round(c_unique * effective_step + effective_min - effective_step, 9)
            else:
                unique_max = c_unique * effective_step + effective_min - effective_step
            result = NRange(effective_min, unique_max, effective_step)

            if result.maxValue is not None and effective_max is not None and result.maxValue > effective_max:
                self.logger.warning("Computed maxValue for column [%s] of %s is greater than specified maxValue %s",
                                    self.name,
                                    result.maxValue,
                                    effective_max)
        elif c_range is not None:
            result = c_range
        elif c_range is None:
            effective_min, effective_max, effective_step = None, None, None
            effective_min = coalesce_values(c_min, 0)
            effective_step = coalesce_values(c_step, 1)
            result = NRange(effective_min, c_max, effective_step)
        else:
            result = NRange(0, None, None)
        # assume numeric range of 0 to x, if no range specified
        self.logger.debug("Computing adjusted range for column: %s - %s", self.name, result)

        return result

    def _computeAdjustedDateTimeRangeForColumn(self, colType, c_begin, c_end, c_interval, c_range, c_unique):
        """Determine adjusted range for Date or Timestamp data column
        """
        effective_begin, effective_end, effective_interval = None, None, None
        if c_range is not None and type(c_range) is DateRange:
            effective_begin = c_range.begin
            effective_end = c_range.end
            effective_interval = c_range.interval
        effective_interval = coalesce_values(effective_interval, c_interval)
        effective_end = coalesce_values(effective_end, c_end)
        effective_begin = coalesce_values(effective_begin, c_begin)

        if type(colType) is DateType:
            result = DateRange.computeDateRange(effective_begin, effective_end, effective_interval, c_unique)
        else:
            result = DateRange.computeTimestampRange(effective_begin, effective_end, effective_interval, c_unique)

        self.logger.debug("Computing adjusted range for column: %s - %s", self.name, result)
        return result

    def _getUniformRandomExpression(self, col_name):
        """ Get random expression accounting for seed method

        :returns: expression of ColDef form - i.e `lit`, `expr` etc

        The value returned will be a number between 0 and 1 inclusive
        """
        assert col_name is not None, "`col_name` must not be None"
        if self._randomSeedMethod == RANDOM_SEED_FIXED and self._randomSeed != RANDOM_SEED_RANDOM:
            return expr(f"rand({self._randomSeed})")
        elif self._randomSeedMethod == RANDOM_SEED_HASH_FIELD_NAME:
            assert self.name is not None, " `self.name` must not be none"
            return expr(f"rand(hash('{self.name}'))")
        else:
            return rand()

    def _getRandomExpressionForDistribution(self, col_name, col_distribution):
        """ Get random expression accounting for seed method

        :returns: expression of ColDef form - i.e `lit`, `expr` etc

        The value returned will be a number between 0 and 1 inclusive
        """
        assert col_name is not None and len(col_name) > 0, "`col_name` must not be None and non empty"
        assert col_distribution is not None, "`col_distribution` must not be None"
        assert isinstance(col_distribution, DataDistribution), \
            "`distribution` object must be an instance of data distribution"

        self.executionHistory.append(f".. random number generation via distribution `{col_distribution}`")

        return col_distribution.generateNormalizedDistributionSample()

    def _getUniformRandomSQLExpression(self, col_name):
        """ Get random SQL expression accounting for seed method

        :returns: expression as a SQL string
        """
        assert col_name is not None, " `col_name` must not be None"
        if self._randomSeedMethod == RANDOM_SEED_FIXED and self._randomSeed != RANDOM_SEED_RANDOM:
            assert self._randomSeed is not None, "`randomSeed` must not be None"
            return f"rand({self._randomSeed})"
        elif self._randomSeedMethod == RANDOM_SEED_HASH_FIELD_NAME:
            assert self.name is not None, "`self.name` must not be none"
            return f"rand(hash('{self.name}'))"
        else:
            return "rand()"

    def _getScaledIntSQLExpression(self, col_name, scale, base_columns, base_datatypes=None, compute_method=None,
                                   normalize=False):
        """ Get scaled numeric expression

        This will produce a scaled SQL expression from the base columns



        :param col_name: = Column name used for error messages and debugging
        :param normalize: = If True, will normalize to the range 0 .. 1 inclusive
        :param scale: = Numeric value indicating scaling factor - will scale via modulo arithmetic
        :param base_columns: = list of base_columns
        :param base_datatypes: = list of Spark SQL datatypes for columns
        :param compute_method: = indicates how the value is be derived from base columns - i.e 'hash' or 'values'
                               - treated as hint only
        :returns: scaled expression as a SQL string

        """
        assert col_name is not None, "`col_name` must not be None"
        assert self.name is not None, "`self.name` must not be None"
        assert scale is not None, "`scale` must not be None"
        assert (compute_method is None or
                compute_method in COMPUTE_METHOD_VALID_VALUES), "`compute_method` must be valid value "
        assert (base_columns is not None and
                type(base_columns) is list
                and len(base_columns) > 0), "Base columns must be a non-empty list"

        effective_compute_method = compute_method

        # if we have multiple columns, effective compute method is always the hash of the base values
        if len(base_columns) > 1:
            if compute_method == VALUES_COMPUTE_METHOD:
                self.logger.warning(
                    "For column generation with values and multiple base columns,  data will  be computed with `hash`")
            effective_compute_method = HASH_COMPUTE_METHOD

        if effective_compute_method is None or effective_compute_method is AUTO_COMPUTE_METHOD:
            effective_compute_method = VALUES_COMPUTE_METHOD

        column_set = ",".join(base_columns)

        if effective_compute_method == HASH_COMPUTE_METHOD:
            result = f"cast( floor((hash({column_set}) % {scale}) + {scale}) % {scale} as double)"
        else:
            result = f"cast( ( floor(({column_set} % {scale}) + {scale}) % {scale}) as double) "

        if normalize:
            result = f"({result} / {(scale * 1.0) - 1.0})"

        self.logger.debug("computing scaled field [%s] as expression [%s]", col_name, result)
        return result

    @property
    def isWeightedValuesColumn(self):
        """ check if column is a weighed values column """
        return self['weights'] is not None and self.values is not None

    def getNames(self):
        """ get column names as list of strings"""
        min_num_columns, max_num_columns, struct_type = self._getMultiColumnDetails(validate=False)

        if max_num_columns > 1 and struct_type is None:
            return [f"{self.name}_{x}" for x in range(0, max_num_columns)]
        else:
            return [self.name]

    def getNamesAndTypes(self):
        """ get column names as list of tuples `(name, datatype)`"""
        min_num_columns, max_num_columns, struct_type = self._getMultiColumnDetails(validate=False)

        if max_num_columns > 1 and struct_type is None:
            return [(f"{self.name}_{x}", self.datatype) for x in range(0, max_num_columns)]
        else:
            return [(self.name, self.datatype)]

    def keys(self):
        """ Get the keys as list of strings """
        assert self._csOptions is not None, "self._csOptions should be non-empty"
        return self._csOptions.keys()

    def __getitem__(self, key):
        """ implement the built in dereference by key behavior """
        assert key is not None, "key should be non-empty"
        return self._csOptions.getOrElse(key, None)

    @property
    def isFieldOmitted(self):
        """ check if this field should be omitted from the output

        If the field is omitted from the output, the field is available for use in expressions etc.
        but dropped from the final set of fields
        """
        return self.omit

    @property
    def baseColumn(self):
        """get the base column used to generate values for this column"""
        return self['baseColumn']

    @property
    def datatype(self):
        """get the Spark SQL data type used to generate values for this column"""
        return self['type']

    @property
    def prefix(self):
        """get the string prefix used to generate values for this column

        When a string field is generated from this spec, the prefix is prepended to the generated string
        """
        return self['prefix']

    @property
    def suffix(self):
        """get the string suffix used to generate values for this column

        When a string field is generated from this spec, the suffix is appended to the generated string
        """
        return self['suffix']

    @property
    def min(self):
        """get the column generation `minValue` value used to generate values for this column"""
        return self._dataRange.minValue

    @property
    def max(self):
        """get the column generation `maxValue` value used to generate values for this column"""
        return self['maxValue']

    @property
    def step(self):
        """get the column generation `step` value used to generate values for this column"""
        return self['step']

    @property
    def exprs(self):
        """get the column generation `exprs` attribute used to generate values for this column.
        """
        return self['exprs']

    @property
    def expr(self):
        """get the `expr` attributed used to generate values for this column"""
        return self['expr']

    @property
    def text_separator(self):
        """get the `expr` attributed used to generate values for this column"""
        return self['text_separator']

    @property
    def begin(self):
        """get the `begin` attribute used to generate values for this column

        For numeric columns, the range (minValue, maxValue, step) is used to control data generation.
        For date and time columns, the range (begin, end, interval) are used to control data generation
        """
        return self['begin']

    @property
    def end(self):
        """get the `end` attribute used to generate values for this column

        For numeric columns, the range (minValue, maxValue, step) is used to control data generation.
        For date and time columns, the range (begin, end, interval) are used to control data generation
        """
        return self['end']

    @property
    def interval(self):
        """get the `interval` attribute used to generate values for this column

        For numeric columns, the range (minValue, maxValue, step) is used to control data generation.
        For date and time columns, the range (begin, end, interval) are used to control data generation
        """
        return self['interval']

    @property
    def numColumns(self):
        """get the `numColumns` attribute used to generate values for this column

        if a column is specified with the `numColumns` attribute, this is used to create multiple
        copies of the column, named `colName1` .. `colNameN`
        """
        return self['numColumns']

    @property
    def numFeatures(self):
        """get the `numFeatures` attribute used to generate values for this column

        if a column is specified with the `numFeatures` attribute, this is used to create multiple
        copies of the column, combined into an array or feature vector
        """
        return self['numFeatures']

    def structType(self):
        """get the `structType` attribute used to generate values for this column

        When a column spec is specified to generate multiple copies of the column, this controls whether
        these are combined into an array etc
        """
        return self['structType']

    def getOrElse(self, key, default=None):
        """ Get value for option key if it exists or else return default

        :param key: key name for option
        :param default: default value if option was not provided
        :return: option value or default

        """
        return self._csOptions.getOrElse(key, default)

    def getPlanEntry(self):
        """ Get execution plan entry for object

        :returns: String representation of plan entry
        """
        desc = self['description']
        if desc is not None:
            return " |-- " + desc
        else:
            return f" |-- building column generator for column {self.name}"

    def _makeWeightedColumnValuesExpression(self, values, weights, seed_column_name):
        """make SQL expression to compute the weighted values expression

        :returns: Spark SQL expr
        """
        from .function_builder import ColumnGeneratorBuilder
        assert values is not None, "`values` expression must be supplied as list of values"
        assert weights is not None, "`weights` expression must be list of weights"
        assert len(values) == len(weights), "`weights` and `values` lists must be of equal length"
        assert seed_column_name is not None, "`seed_column_name` must be explicit column name"
        expr_str = ColumnGeneratorBuilder.mkExprChoicesFn(values, weights, seed_column_name, self.datatype)
        return expr(expr_str).astype(self.datatype)

    def _isRealValuedColumn(self):
        """ determine if column is real valued

        :returns: Boolean - True if condition is true
        """
        col_type_name = self['type'].typeName()

        return col_type_name in ['double', 'float', 'decimal']

    def _isDecimalColumn(self):
        """ determine if column is decimal column

        :returns: Boolean - True if condition is true
        """
        col_type_name = self['type'].typeName()

        return col_type_name == 'decimal'

    def _isContinuousValuedColumn(self):
        """ determine if column generates continuous values

        :returns: Boolean - True if condition is true
        """
        is_continuous = self['continuous']

        return is_continuous

    def _getSeedExpression(self, base_column):
        """ Get seed expression for column generation

        This is used to generate the base value for every column
        if using a single base column, then simply use that, otherwise use either
        a SQL hash of multiple columns, or an array of the base column values converted to strings

        :returns: Spark SQL `col` or `expr` object
        """

        if type(base_column) is list:
            assert len(base_column) > 0, "`baseColumn` must be list of column names"
            if len(base_column) == 1:
                if self._baseColumnComputeMethod == HASH_COMPUTE_METHOD:
                    return expr(f"hash({base_column[0]})")
                else:
                    return col(base_column[0])
            elif self._baseColumnComputeMethod == VALUES_COMPUTE_METHOD:
                base_values = [f"string(ifnull(`{x}`, 'null'))" for x in base_column]
                return expr(f"array({','.join(base_values)})")
            else:
                return expr(f"hash({','.join(base_column)})")
        else:
            if self._baseColumnComputeMethod == HASH_COMPUTE_METHOD:
                return expr(f"hash({base_column})")
            else:
                return col(base_column)

    def _isStringField(self):
        return type(self.datatype) is StringType

    def _computeRangedColumn(self, datarange, base_column, is_random):
        """ compute a ranged column

        maxValue is maxValue actual value

        :returns: spark sql `column` or expression that can be used to generate a column
        """
        assert base_column is not None, "`baseColumn` must be specified"
        assert datarange is not None, "`datarange` must be specified"
        assert datarange.isFullyPopulated(), "`datarange` must be fully populated (minValue, maxValue, step)"

        if is_random:
            if self.distribution is not None:
                random_generator = self._getRandomExpressionForDistribution(self.name, self.distribution)
            else:
                random_generator = self._getUniformRandomExpression(self.name)
        else:
            random_generator = None

        if self._isContinuousValuedColumn() and self._isRealValuedColumn() and is_random:
            crange = datarange.getContinuousRange()
            baseval = random_generator * lit(crange)
        else:
            crange = datarange.getDiscreteRange()
            modulo_factor = lit(crange + 1)
            # following expression is needed as spark sql modulo of negative number is negative
            modulo_exp = ((self._getSeedExpression(base_column) % modulo_factor) + modulo_factor) % modulo_factor
            baseval = (modulo_exp * lit(datarange.step)) if not is_random else (
                    sql_round(random_generator * lit(crange)) * lit(datarange.step))

        if self._baseColumnComputeMethod == VALUES_COMPUTE_METHOD:
            new_def = self._adjustForMinValue(baseval, datarange)
        elif self._baseColumnComputeMethod == RAW_VALUES_COMPUTE_METHOD:
            new_def = baseval
        else:
            new_def = self._adjustForMinValue(baseval, datarange, force=True)

        # for ranged values in strings, use type of minValue, maxValue and step as output type
        if type(self.datatype) is StringType:
            if type(datarange.min) is float or type(datarange.max) is float or type(datarange.step) is float:
                if datarange.getScale() > 0:
                    new_def = sql_round(new_def.astype(FloatType()), datarange.getScale())
                else:
                    new_def = new_def.astype(DoubleType())
            else:
                new_def = new_def.astype(IntegerType())

        return new_def

    def _adjustForMinValue(self, baseval, datarange, force=False):
        """ Adjust for minimum value of data range
        :param baseval: base expression
        :param datarange: data range to conform to
        :param force: always adjust (possibly for implicit cast reasons)
        """
        if force and datarange is not None:
            new_def = baseval + lit(datarange.minValue)
        elif (datarange is not None) and (datarange.minValue != 0) and (datarange.minValue != 0.0):
            new_def = baseval + lit(datarange.minValue)
        else:
            new_def = baseval
        return new_def

    def _makeSingleGenerationExpression(self, index=None, use_pandas_optimizations=True):
        """ generate column data for a single column value via Spark SQL expression

            :param index: for multi column generation, specifies index of column being generated
            :param use_pandas_optimizations: if True, uses Pandas vectorized optimizations. Defaults to `True`
            :returns: spark sql `column` or expression that can be used to generate a column
        """
        self.logger.debug("building column : %s", self.name)

        # get key column specification properties
        col_is_rand, cdistribution = self[OPTION_RANDOM], self['distribution']
        percent_nulls = self['percentNulls']
        sformat = self['format']

        if self._dataRange is not None:
            self._dataRange.adjustForColumnDatatype(self.datatype)
            self.executionHistory.append(f".. using effective range: {self._dataRange}")

        new_def = None

        # generate expression
        if type(self.datatype) in [ArrayType, MapType, StructType] and self.expr is None:
            self.logger.warning("Array, Map or Struct type column with no SQL `expr` will result in NULL value")
            self.executionHistory.append(".. WARNING: Array, Map or Struct type column with no SQL `expr` ")

        # handle weighted values for weighted value columns
        # a weighted values column will use a base value denoted by `self._weightedBaseColumn`
        if self.isWeightedValuesColumn:
            self.executionHistory.append(".. building weighted volumn values expression")
            new_def = self._makeWeightedColumnValuesExpression(self.values, self.weights, self._weightedBaseColumn)

            if type(self.datatype) is StringType and self.textGenerator is not None:
                self.logger.warning("Template generation / text generation not supported for weighted columns")
                self.executionHistory.append(".. WARNING: Template & text generation not supported for weights")

            if type(self.datatype) is StringType and sformat is not None:
                self.logger.warning("Formatting not supported for weighted columns")
                self.executionHistory.append(".. WARNING: Formatting not supported for weighted columns")
        else:
            # rs: initialize the begin, end and interval if not initialized for date computations
            # defaults are start of day, now, and 1 minute respectively

            # for array, struct and map types, either value is provided via `expr` or via values
            if not type(self.datatype) in [ArrayType, MapType, StructType] or self.values is not None:
                self._computeImpliedRangeIfNeeded(self.datatype)

            # TODO: add full support for date value generation
            if self.expr is not None:
                # note use of SQL expression ignores range specifications
                new_def = expr(self.expr)
                self.executionHistory.append(f".. using SQL expression `{self.expr}` as base")

                if not self._inferDataType:
                    new_def = new_def.astype(self.datatype)
                    self.executionHistory.append(f".. casting to  `{self.datatype}`")
            elif type(self.datatype) in [ArrayType, MapType, StructType] and self.values is None:
                new_def = expr("NULL")
            elif self._dataRange is not None and self._dataRange.isFullyPopulated():
                self.executionHistory.append(f".. computing ranged value: {self._dataRange}")
                new_def = self._computeRangedColumn(base_column=self.baseColumn, datarange=self._dataRange,
                                                    is_random=col_is_rand)
            elif type(self.datatype) is DateType:
                # TODO: fixup for date generation

                # record execution history
                self.executionHistory.append(".. using random date expression")
                sql_random_generator = self._getUniformRandomSQLExpression(self.name)
                new_def = expr(f"date_sub(current_date, rounding({sql_random_generator}*1024))").astype(
                    self.datatype)
            else:
                if self._baseColumnComputeMethod == VALUES_COMPUTE_METHOD:
                    self.executionHistory.append(".. using values compute expression for seed")
                    new_def = self._getSeedExpression(self.baseColumn)
                elif self._baseColumnComputeMethod == RAW_VALUES_COMPUTE_METHOD:
                    self.executionHistory.append(".. using raw values compute expression for seed")
                    new_def = self._getSeedExpression(self.baseColumn)
                # TODO: resolve issues with hash when using templates
                # elif self._baseColumnComputeMethod == HASH_COMPUTE_METHOD:
                #    newDef = self._getSeedExpression(self.baseColumn)
                else:
                    self.logger.info("Assuming a seeded base expression with minimum value for column %s", self.name)
                    self.executionHistory.append(f".. seeding with minimum `{self._dataRange.minValue}`")
                    new_def = ((self._getSeedExpression(self.baseColumn) + lit(self._dataRange.minValue))
                               .astype(self.datatype))

            if self.values is not None:
                new_def = F.element_at(F.array([F.lit(x) for x in self.values]), new_def.astype(IntegerType()) + 1)
            elif type(self.datatype) is StringType and self.expr is None:
                new_def = self._applyPrefixSuffixExpressions(self.prefix, self.suffix, new_def)

            # use string generation template if available passing in what was generated to date
            if type(self.datatype) is StringType and self.textGenerator is not None:
                new_def = self._applyTextGenerationExpression(new_def, use_pandas_optimizations)

        if type(self.datatype) is StringType and sformat is not None:
            new_def = self._applyTextFormatExpression(new_def, sformat)

        new_def = self._applyFinalCastExpression(self.datatype, new_def)

        if percent_nulls is not None:
            new_def = self._applyComputePercentNullsExpression(new_def, percent_nulls)
        return new_def

    def _onSelect(self, df):
        """
        The _onSelect method is called when the column specifications expression as produced by the
        method ``_makeSingleGenerationExpression`` is used in a select statement.

        :param df: Dataframe in which expression is used
        :return: nothing

        .. note:: The purpose of this method is to allow for introspection of information such as datatype
                  which can only be determined when column specifications expression is used.
        """
        if self._inferDataType:
            inferred_type = df.schema[self.name].dataType
            self.logger.info("Inferred datatype for column %s as %s", self.name, str(inferred_type))
            self._csOptions.options['type'] = inferred_type

    def _applyTextFormatExpression(self, new_def, sformat):
        # note :
        # while it seems like this could use a shared instance, this does not work if initialized
        # in a class method
        self.executionHistory.append(f".. applying column format  `{sformat}`")
        new_def = format_string(sformat, new_def)
        return new_def

    def _applyPrefixSuffixExpressions(self, cprefix, csuffix, new_def):
        # string value generation is simply handled by combining with a suffix or prefix
        # TODO: prefix and suffix only apply to base columns that are numeric types
        text_separator = self.text_separator if self.text_separator is not None else '_'
        if cprefix is not None and csuffix is not None:
            self.executionHistory.append(".. applying column prefix and suffix")
            new_def = concat(lit(cprefix), lit(text_separator), new_def.astype(IntegerType()), lit(text_separator),
                             lit(csuffix))
        elif cprefix is not None:
            self.executionHistory.append(".. applying column prefix")
            new_def = concat(lit(cprefix), lit(text_separator), new_def.astype(IntegerType()))
        elif csuffix is not None:
            self.executionHistory.append(".. applying column suffix")
            new_def = concat(new_def.astype(IntegerType()), lit(text_separator), lit(csuffix))
        return new_def

    def _applyTextGenerationExpression(self, new_def, use_pandas_optimizations):
        """Apply text generation expression to column expression

        :param new_def : column definition being created
        :param use_pandas_optimizations: Whether Pandas optimizations should be applied
        :returns: new column definition
        """
        # note :
        # while it seems like this could use a shared instance, this does not work if initialized
        # in a class method
        tg = self.textGenerator
        if use_pandas_optimizations:
            self.executionHistory.append(f".. text generation via pandas scalar udf `{tg}`")
            u_value_from_generator = pandas_udf(tg.pandasGenerateText,
                                                returnType=StringType()).asNondeterministic()
        else:
            self.executionHistory.append(f".. text generation via udf `{tg}`")
            u_value_from_generator = udf(tg.classicGenerateText,
                                         StringType()).asNondeterministic()
        new_def = u_value_from_generator(new_def)
        return new_def

    def _applyFinalCastExpression(self, col_type, new_def):
        """ Apply final cast expression for column data

        :param col_type: final column type
        :param new_def:  column definition being created
        :returns: new column definition
        """
        self.executionHistory.append(f".. casting column [{self.name}] to  `{col_type}`")

        # cast the result to the appropriate type. For dates, cast first to timestamp, then to date
        if type(col_type) is DateType:
            new_def = new_def.astype(TimestampType()).astype(col_type)
        elif self._inferDataType:
            # dont apply cast when column has an inferred data type
            pass
        else:
            new_def = new_def.astype(col_type)

        return new_def

    def _applyComputePercentNullsExpression(self, newDef, probabilityNulls):
        """Compute percentage nulls for column being generated

           :param newDef: Column definition being created
           :param probabilityNulls: Probability of nulls to be generated for particular column. Values can be 0.0 - 1.0
           :returns: new column definition with probability of nulls applied
        """
        assert self.nullable, f"Column `{self.name}` must be nullable for `percent_nulls` option"
        self.executionHistory.append(".. applying null generator - `when rnd > prob then value - else null`")

        assert probabilityNulls is not None, "option 'percent_nulls' must not be null value or None"
        assert type(probabilityNulls) in [int, float], "option 'percent_nulls' must be int or float"
        assert 0.0 <= probabilityNulls <= 1.0, "option 'percent_nulls' must in the range [0.0 .. 1.0]"
        prob_nulls = probabilityNulls * 1.0  # for edge case where int was passed
        random_generator = self._getUniformRandomExpression(self.name)
        newDef = when(random_generator > lit(prob_nulls), newDef).otherwise(lit(None))
        return newDef

    def _computeImpliedRangeIfNeeded(self, col_type):
        """ Compute implied range if necessary
            :param col_type" Column type
            :returns: nothing
        """
        # check for implied ranges
        if self.values is not None:
            self._dataRange = NRange(0, len(self.values) - 1, 1)
        elif type(col_type) is BooleanType:
            self._dataRange = NRange(0, 1, 1)
        self.executionHistory.append(f".. using adjusted effective range: {self._dataRange}")

    def _getMultiColumnDetails(self, validate):
        """ Determine min and max number of columns to generate along with `structType` for columns
            with multiple columns / features

        :param validate:  If true, raises ValueError if there are bad option entries
        :return: tuple of (min_columns, max_columns, structType
        """
        num_columns = self['numColumns']
        struct_type = self['structType']

        if num_columns is None:
            num_columns = self['numFeatures']

        if num_columns is None:
            min_num_columns, max_num_columns = 1, 1
        elif isinstance(num_columns, int):
            min_num_columns, max_num_columns = int(num_columns), int(num_columns)
        elif isinstance(num_columns, tuple):
            if validate and ((len(num_columns) != 2) or not all(isinstance(c, int) for c in num_columns)):
                raise ValueError(f"Bad value [{num_columns}] for `numColumns` / `numFeatures` attribute")

            min_num_columns, max_num_columns = int(num_columns[0]), int(num_columns[1])

            if validate and (min_num_columns > max_num_columns):
                raise ValueError(f"Bad value [{num_columns}] for `numColumns` / `numFeatures` attribute")
        else:
            if validate:
                raise ValueError(f"Bad value [{num_columns}] for `numColumns` / `numFeatures` attribute")

            min_num_columns, max_num_columns = 1, 1

        if validate and (min_num_columns != max_num_columns) and (struct_type != self._ARRAY_STRUCT_TYPE):
            self.logger.warning(
                f"Varying number of features / columns specified for non-array column [{self.name}]")
            self.logger.warning(
                f"Lower bound for number of features / columns ignored for [{self.name}]")
            min_num_columns = max_num_columns

        return min_num_columns, max_num_columns, struct_type

    def makeGenerationExpressions(self):
        """ Generate structured column if multiple columns or features are specified

        if there are multiple columns / features specified using a single definition, it will generate
        a set of columns conforming to the same definition,
        renaming them as appropriate and combine them into a array if necessary
        (depending on the structure combination instructions)

            :param self: is ColumnGenerationSpec for column
            :returns: spark sql `column` or expression that can be used to generate a column
        """
        min_num_columns, max_num_columns, struct_type = self._getMultiColumnDetails(validate=True)

        self.executionHistory = []

        if (min_num_columns == 1) and (max_num_columns == 1) and struct_type != self._ARRAY_STRUCT_TYPE:
            # record execution history for troubleshooting
            self.executionHistory.append(f"generating single column - `{self.name}` having type `{self.datatype}`")

            retval = self._makeSingleGenerationExpression(use_pandas_optimizations=True)

            # record how column was generated
            exec_step_history = ".. computed from base values - "
            exec_step_history += f"`{self.baseColumn}`, method: `{self._baseColumnComputeMethod}`"
            self.executionHistory.append(exec_step_history)
        else:
            self.executionHistory.append(f"generating multiple columns {max_num_columns} - `{self.name}`")

            retval = []

            for ix in range(max_num_columns):
                with self._temporaryRename(f"{self.name}_{ix}") as renamed_cs:
                    retval.append(renamed_cs._makeSingleGenerationExpression(ix, use_pandas_optimizations=True))

            if struct_type == 'array':
                self.executionHistory.append(".. converting multiple columns to array")
                retval = array(retval)

                if min_num_columns != max_num_columns:
                    column_set = ",".join(self.baseColumns)
                    diff = max_num_columns - min_num_columns
                    expr_str = f"{min_num_columns} + (abs(hash({column_set})) % {diff + 1})"

                    retval = F.slice(retval, F.lit(1), F.expr(expr_str))

        return retval



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This module defines the package contents for the test data generator library

The main entry point for any test data generation activities is
the `DataGenerator` class.

Most of the other classes are used for internal purposes only
"""

from .data_generator import DataGenerator
from .datagen_constants import DEFAULT_RANDOM_SEED, RANDOM_SEED_RANDOM, RANDOM_SEED_FIXED, \
                               RANDOM_SEED_HASH_FIELD_NAME, MIN_PYTHON_VERSION, MIN_SPARK_VERSION, \
                               INFER_DATATYPE, SPARK_DEFAULT_PARALLELISM
from .utils import ensure, topologicalSort, mkBoundsList, coalesce_values, \
    deprecated, parse_time_interval, DataGenError, split_list_matching_condition, strip_margins, \
    json_value_from_path, system_time_millis
from ._version import __version__
from .column_generation_spec import ColumnGenerationSpec
from .column_spec_options import ColumnSpecOptions
from .data_analyzer import DataAnalyzer
from .schema_parser import SchemaParser
from .daterange import DateRange
from .datarange import DataRange
from .nrange import NRange
from .function_builder import ColumnGeneratorBuilder
from .spark_singleton import SparkSingleton
from .text_generators import TemplateGenerator, ILText, TextGenerator
from .text_generator_plugins import PyfuncText, PyfuncTextFactory, FakerTextFactory, fakerText
from .html_utils import HtmlUtils

__all__ = ["data_generator", "data_analyzer", "schema_parser", "daterange", "nrange",
           "column_generation_spec", "utils", "function_builder",
           "spark_singleton", "text_generators", "datarange", "datagen_constants",
           "text_generator_plugins", "html_utils"
           ]


def python_version_check(python_version_expected):
    """Check against Python version

       Allows minimum version to be passed in to facilitate unit testing

       :param python_version_expected: = minimum version of python to support as tuple e.g (3,6)
       :return: True if passed

        """
    import sys
    return sys.version_info >= python_version_expected


# lets check for a correct python version or raise an exception
if not python_version_check(MIN_PYTHON_VERSION):
    raise RuntimeError(f"Minimum version of Python supported is {MIN_PYTHON_VERSION[0]}.{MIN_PYTHON_VERSION[1]}")



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This module defines the `NRange` class used to specify data ranges
"""

import math

from pyspark.sql.types import LongType, FloatType, IntegerType, DoubleType, ShortType, \
    ByteType

from .datarange import DataRange

_OLD_MIN_OPTION = 'min'
_OLD_MAX_OPTION = 'max'


class NRange(DataRange):
    """ Ranged numeric interval representing the interval minValue .. maxValue inclusive

    A ranged object can be uses as an alternative to the `minValue`, `maxValue`, `step` parameters
    to the DataGenerator `withColumn` and `withColumn` objects.
    Specify by passing an instance of `NRange` to the `dataRange` parameter.

    :param minValue: Minimum value of range. May be integer / long / float
    :param maxValue: Maximum value of range. May be integer / long / float
    :param step: Step value for range. May be integer / long / float
    :param until: Upper bound for range ( i.e maxValue+1)

    You may only specify a `maxValue` or `until` value not both.

    For a decreasing sequence, use a negative step value.
    """

    def __init__(self, minValue=None, maxValue=None, step=None, until=None, **kwArgs):
        # check if older form of `minValue` and `maxValue` are used, and if so
        if _OLD_MIN_OPTION in kwArgs:
            assert minValue is None, \
                "Only one of `minValue` and `minValue` can be specified. Use of `minValue` is preferred"
            self.minValue = kwArgs[_OLD_MIN_OPTION]
            kwArgs.pop(_OLD_MIN_OPTION, None)
        else:
            self.minValue = minValue

        if _OLD_MAX_OPTION in kwArgs:
            assert maxValue is None, \
                "Only one of `maxValue` and `maxValue` can be specified. Use of `maxValue` is preferred"
            self.maxValue = kwArgs[_OLD_MAX_OPTION]
            kwArgs.pop(_OLD_MAX_OPTION, None)
        else:
            self.maxValue = maxValue
        assert len(kwArgs.keys()) == 0, "no keyword options other than `min` and `max` allowed"

        assert until is None if self.maxValue is not None else True, "Only one of maxValue or until can be specified"
        assert self.maxValue is None if until is not None else True, "Only one of maxValue or until can be specified"

        if until is not None:
            self.maxValue = until + 1
        self.step = step

    def __str__(self):
        return f"NRange({self.minValue}, {self.maxValue}, {self.step})"

    def isEmpty(self):
        """Check if object is empty (i.e all instance vars of note are `None`

        :returns: `True` if empty, `False` otherwise
        """
        return self.minValue is None and self.maxValue is None and self.step is None

    def isFullyPopulated(self):
        """Check is all instance vars are populated

        :returns: `True` if fully populated, `False` otherwise
        """
        return self.minValue is not None and self.maxValue is not None and self.step is not None

    def adjustForColumnDatatype(self, ctype):
        """ Adjust default values for column output type

        :param ctype: Spark SQL type instance to adjust range for
        :returns: No return value - executes for effect only
        """
        if ctype.typeName() == 'decimal':
            if self.minValue is None:
                self.minValue = 0.0
            if self.maxValue is None:
                self.maxValue = math.pow(10, ctype.precision - ctype.scale) - 1.0
            if self.step is None:
                self.step = 1.0

        if type(ctype) is ShortType and self.maxValue is not None:
            assert self.maxValue <= 65536, "`maxValue` must be in range of short"

        if type(ctype) is ByteType and self.maxValue is not None:
            assert self.maxValue <= 256, "`maxValue` must be in range of byte (0 - 256)"

        if (type(ctype) is DoubleType or type(ctype) is FloatType) and self.step is None:
            self.step = 1.0

        if (type(ctype) is ByteType
              or type(ctype) is ShortType
              or type(ctype) is IntegerType
              or type(ctype) is LongType) and self.step is None:
            self.step = 1

    def getDiscreteRange(self):
        """Convert range to discrete range

        :returns: number of discrete values in range. For example `NRange(1, 5, 0.5)` has 8 discrete values

        .. note::
           A range of 0,4, 0.5 has 8 discrete values not 9 as the `maxValue` value is not part of the range

        TODO: check range of values

        """
        if type(self.minValue) is int and type(self.maxValue) is int and self.step == 1:
            return self.maxValue - self.minValue
        else:
            # when any component is a float, we will return a float for the discrete range
            # to simplify computations
            return float(math.floor((self.maxValue - self.minValue) * float(1.0 / self.step)))

    def getContinuousRange(self):
        """Convert range to continuous range

        :returns: float value for size of interval from `minValue` to `maxValue`
        """
        return (self.maxValue - self.minValue) * float(1.0)

    def getScale(self):
        """Get scale of range"""
        smin, smax, sstep = 0, 0, 0

        if self.minValue is not None:
            smin = self._precision_and_scale(self.minValue)[1]
        if self.maxValue is not None:
            smax = self._precision_and_scale(self.maxValue)[1]
        if self.step is not None:
            sstep = self._precision_and_scale(self.step)[1]

        # return maximum scale of components
        return max(smin, smax, sstep)

    def _precision_and_scale(self, x):
        max_digits = 14
        int_part = int(abs(x))
        magnitude = 1 if int_part == 0 else int(math.log10(int_part)) + 1
        if magnitude >= max_digits:
            return (magnitude, 0)
        frac_part = abs(x) - int_part
        multiplier = 10 ** (max_digits - magnitude)
        frac_digits = multiplier + int(multiplier * frac_part + 0.5)
        while frac_digits % 10 == 0:
            frac_digits /= 10
        scale = int(math.log10(frac_digits))
        return (magnitude + scale, scale)



"""A setuptools based setup module
"""
import setuptools
import re

START_TAG=r"^\s*\<\!--.*exclude\s+package.*--\>\s*$"
END_TAG=r"^\s*\<\!--.*end\s+exclude\s+package.*--\>\s*$"

with open("README.md", "r") as fh:
    # exclude lines from readme that dont apply to publication in package
    # for example navigation bar in Readme refers to github relative paths
    long_description = fh.read()
    modified_description_lines = []
    marked = False

    for line in long_description.split("\n"):
        if re.match(START_TAG, line) or re.match(END_TAG, line):
            marked = True
        if not marked:
            modified_description_lines.append(line)
        if  re.match(END_TAG, line):
            marked = False

    long_description = "\n".join(modified_description_lines)

package_long_description = """###Databricks Labs Spark Data Generator###

    The Databricks Labs Data Generator can generate realistic synthetic data sets at scale for testing, 
    benchmarking, environment validation and other purposes.
    """

setuptools.setup(
    name="dbldatagen",
    version="0.3.6",
    author="Ronan Stokes, Databricks",
    description="Databricks Labs -  PySpark Synthetic Data Generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/databrickslabs/data-generator",
    project_urls={
        "Databricks Labs": "https://www.databricks.com/learn/labs",
        "Documentation": "https://databrickslabs.github.io/dbldatagen/public_docs/index.html"
},
    packages=['dbldatagen',
              'dbldatagen.distributions'
              ],
    license="Databricks License",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Testing",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators"
    ],
    python_requires='>=3.8.10',
)



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file defines the `DataGenError` classes and utility functions

These are meant for internal use only
"""

import functools
import json
import re
import time
import warnings
from datetime import timedelta

import jmespath


def deprecated(message=""):
    """
    Define a deprecated decorator without dependencies on 3rd party libraries

    Note there is a 3rd party library called `deprecated` that provides this feature but goal is to only have
    dependencies on packages already used in the Databricks runtime
    """

    # create closure around function that follows use of the decorator

    def deprecated_decorator(func):
        @functools.wraps(func)
        def deprecated_func(*args, **kwargs):
            warnings.warn(f"`{func.__name__}` is a deprecated function or method. \n{message}",
                          category=DeprecationWarning, stacklevel=1)
            warnings.simplefilter('default', DeprecationWarning)
            return func(*args, **kwargs)

        return deprecated_func

    return deprecated_decorator


class DataGenError(Exception):
    """Exception class to represent data generation errors

        :param msg: message related to error
        :param baseException: underlying exception, if any that caused the issue
    """

    def __init__(self, msg, baseException=None):
        """ constructor
        """
        super().__init__(msg)
        self._underlyingException = baseException
        self._msg = msg

    def __repr__(self):
        return f"DataGenError(msg='{self._msg}', baseException={self._underlyingException})"

    def __str__(self):
        return f"DataGenError(msg='{self._msg}', baseException={self._underlyingException})"


def coalesce_values(*args):
    """For a supplied list of arguments, returns the first argument that does not have the value `None`

    :param args: variable list of arguments which are evaluated
    :returns: First argument in list that evaluates to a non-`None` value
    """
    for x in args:
        if x is not None:
            return x
    return None


def ensure(cond, msg="condition does not hold true"):
    """ensure(cond, s) => throws Exception(s) if c is not true

    :param cond: condition to test
    :param msg: Message to add to exception if exception is raised
    :raises: `DataGenError` exception if condition does not hold true
    :returns: Does not return anything but raises exception if condition does not hold
    """

    def strip_margin(text):
        return re.sub(r'\n[ \t]*\|', '\n', text)

    if not cond:
        raise DataGenError(strip_margin(msg))


def mkBoundsList(x, default):
    """ make a bounds list from supplied parameter - otherwise use default

        :param x: integer or list of 2 values that define bounds list
        :param default: default value if X is `None`
        :returns: list of form [x,y]
    """
    if x is None:
        retval = (True, [default, default]) if type(default) is int else (True, list(default))
        return retval
    elif type(x) is int:
        bounds_list = [x, x]
        assert len(bounds_list) == 2, "bounds list must be of length 2"
        return False, bounds_list
    else:
        bounds_list = list(x)
        assert len(bounds_list) == 2, "bounds list must be of length 2"
        return False, bounds_list


def topologicalSort(sources, initial_columns=None, flatten=True):
    """ Perform a topological sort over sources

    Used to compute the column test data generation order of the column generation dependencies.

    The column generation dependencies are based on the value of the `baseColumn` attribute for `withColumn` or
    `withColumnSpec` statements in the data generator specification.

    :arg sources: list of ``(name, set(names of dependencies))`` pairs
    :arg initial_columns: force ``initial_columns`` to be computed first
    :arg flatten: if true, flatten output list
    :returns: list of names in dependency order separated into build phases

    .. note::
       The algorith will give preference to retaining order of inbound sequence
       over modifying order to produce a lower number of build phases.

       Overall the effect is that the input build order should be retained unless there are forward references
    """
    # generate a copy so that we can modify in place
    pending = [(name, set(deps)) for name, deps in sources]
    provided = [] if initial_columns is None else initial_columns[:]
    build_orders = [] if initial_columns is None else [initial_columns]

    while pending:
        next_pending = []
        gen = []
        value_emitted = False
        defer_emitted = False
        gen_provided = []
        for entry in pending:
            name, deps = entry
            deps.difference_update(provided)
            if deps:
                next_pending.append((name, set(deps)))

                # if dependencies will be satisfied by item emitted in this round, defer output
                if not deps.difference(gen_provided):
                    defer_emitted = True
            elif defer_emitted:
                next_pending.append((name, set(deps)))
            elif name in provided:
                value_emitted = True
            else:
                gen.append(name)
                gen_provided.append(name)
                value_emitted = True
        provided.extend(gen_provided)
        build_orders.append(gen)

        if not value_emitted:
            raise ValueError(f"cyclic or missing dependency detected [{next_pending}]")

        pending = next_pending

    if flatten:
        flattened_list = [item for sublist in build_orders for item in sublist]
        return flattened_list
    else:
        return build_orders


PATTERN_NAME_EQUALS_VALUE = re.compile(r"(\w+)\s*\=\s*([0-9]+)")
PATTERN_VALUE_SPACE_NAME = re.compile(r"([0-9]+)\s+(\w+)")
_WEEKS_PER_YEAR = 52


def parse_time_interval(spec):
    """parse time interval from string"""
    hours = 0
    minutes = 0
    weeks = 0
    microseconds = 0
    milliseconds = 0
    seconds = 0
    years = 0
    days = 0

    assert spec is not None, "Must have valid time interval specification"

    # get time specs such as 12 days, etc. Supported timespans are years, days, hours, minutes, seconds
    timespecs = [x.strip() for x in spec.strip().split(",")]

    for ts in timespecs:
        # allow both 'days=1' and '1 day' syntax
        timespec_parts = re.findall(PATTERN_NAME_EQUALS_VALUE, ts)
        # findall returns list of tuples
        if timespec_parts is not None and len(timespec_parts) > 0:
            num_parts = len(timespec_parts[0])
            assert num_parts >= 1, "must have numeric specification and time element such as `12 hours` or `hours=12`"
            time_value = int(timespec_parts[0][num_parts - 1])
            time_type = timespec_parts[0][0].lower()
        else:
            timespec_parts = re.findall(PATTERN_VALUE_SPACE_NAME, ts)
            num_parts = len(timespec_parts[0])
            assert num_parts >= 1, "must have numeric specification and time element such as `12 hours` or `hours=12`"
            time_value = int(timespec_parts[0][0])
            time_type = timespec_parts[0][num_parts - 1].lower()

        if time_type in ["years", "year"]:
            years = time_value
        elif time_type in ["weeks", "weeks"]:
            weeks = time_value
        elif time_type in ["days", "day"]:
            days = time_value
        elif time_type in ["hours", "hour"]:
            hours = time_value
        elif time_type in ["minutes", "minute"]:
            minutes = time_value
        elif time_type in ["seconds", "second"]:
            seconds = time_value
        elif time_type in ["microseconds", "microsecond"]:
            microseconds = time_value
        elif time_type in ["milliseconds", "millisecond"]:
            milliseconds = time_value

    delta = timedelta(
        days=days,
        seconds=seconds,
        microseconds=microseconds,
        milliseconds=milliseconds,
        minutes=minutes,
        hours=hours,
        weeks=weeks + (years * _WEEKS_PER_YEAR)
    )

    return delta


def strip_margins(s, marginChar):
    """
    Python equivalent of Scala stripMargins method
    
    Takes a string (potentially multiline) and strips all chars up and including the first occurrence of `marginChar`.
    Used to control the formatting of generated text

    `strip_margins("one\n    |two\n    |three", '|')`

    will produce 
    
    ``
    one 
    two
    three
    ``

    :param s: string to strip margins from
    :param marginChar: character to strip 
    :return: modified string
    """
    assert s is not None and type(s) is str
    assert marginChar is not None and type(marginChar) is str

    lines = s.split('\n')
    revised_lines = []

    for line in lines:
        if marginChar in line:
            revised_line = line[line.index(marginChar) + 1:]
            revised_lines.append(revised_line)
        else:
            revised_lines.append(line)

    return '\n'.join(revised_lines)


def split_list_matching_condition(lst, cond):
    """
    Split a list on elements that match a condition

    This will find all matches of a specific condition in the list and split the list into sub lists around the
    element that matches this condition.

    It will handle multiple matches performing splits on each match.

    For example, the following code will produce the results below:

    x = ['id', 'city_name', 'id', 'city_id', 'city_pop', 'id', 'city_id', 'city_pop','city_id', 'city_pop','id']
    splitListOnCondition(x, lambda el: el == 'id')


    Result:
    `[['id'], ['city_name'], ['id'], ['city_id', 'city_pop'],
      ['id'], ['city_id', 'city_pop', 'city_id', 'city_pop'], ['id']]`

    :arg lst: list of items to perform condition matches against
    :arg cond: lambda function or function taking single argument and returning True or False
    :returns: list of sublists
    """
    retval = []

    def match_condition(matchList, matchFn):
        """Return first index of element of list matching condition"""
        if matchList is None or len(matchList) == 0:
            return -1

        for i, matchValue in enumerate(matchList):
            if matchFn(matchValue):
                return i

        return -1

    if lst is None:
        retval = lst
    elif len(lst) == 1:
        retval = [lst]
    else:
        ix = match_condition(lst, cond)
        if ix != -1:
            retval.extend(split_list_matching_condition(lst[0:ix], cond))
            retval.append(lst[ix:ix + 1])
            retval.extend(split_list_matching_condition(lst[ix + 1:], cond))
        else:
            retval = [lst]

    # filter out empty lists
    return [el for el in retval if el != []]


def json_value_from_path(searchPath, jsonData, defaultValue):
    """ Get JSON value from JSON data referenced by searchPath

    searchPath should be a JSON path as supported by the `jmespath` package
    (see https://jmespath.org/)

    :param searchPath: A `jmespath` compatible JSON search path
    :param jsonData: The json data to search (string representation of the JSON data)
    :param defaultValue: The default value to be returned if the value was not found
    :return: Returns the json value if present, otherwise returns the default value
    """
    assert searchPath is not None and len(searchPath) > 0, "search path cannot be empty"
    assert jsonData is not None and len(jsonData) > 0, "JSON data cannot be empty"

    jsonDict = json.loads(jsonData)

    jsonValue = jmespath.search(searchPath, jsonDict)

    if jsonValue is not None:
        return jsonValue

    return defaultValue


def system_time_millis():
    """ return system time as milliseconds since start of epoch

    :return: system time millis as long
    """
    curr_time = round(time.time() / 1000)
    return curr_time
